/*
VocaTest is intended as a small and simple tool for testing vocabularies
Copyright (C) 2003, 2004  Olof Holmgren

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <iostream>
#include <string>
#include <deque>
#include "StringHandling.h" 
#include "FileHandling.h"
#include "VocPair.h"
#include "VocList.h"
using namespace std;

// The helpfile :-)
string helpString =
"\nAvailable commands:"
"\n$abort             - Abort currently running test"
"\nappend [file]      - Appends a new vocfile to the current test"
"\nbuildvoc [file]    - Create a small example vocfile"
"\nhelp               - Display this help"
"\nlist               - List the currently loaded words"
"\nload [file]        - Unloads all currently loaded vocfiles and loads a new one"
"\nrandomize          - All following tests will be performed in random order"
"\nsave [file]        - Saves the currently loaded vocfile(s) to a new file"
"\ntest (strict)      - Perform a test on the loaded vocfile(s)"
"\ntestbackw (strict) - Perform a \"backward\" test on the loaded vocfile(s)"
"\nquit, exit         - Quit VocaTest";

// The sample vocfile
string sampleFile =
"// An example of a vocfile for VocaTest\n"
"\n"
"Anything that is not a valid voc will be ignored\n"
"This means that VocaTest will (as far as I can see) never refuse to load a\n"
"(text) file. It will just skip all lines it doesn't understand.\n"
"\n"
"// Everything after // is a comment. Use these to add comments on the\n"
"// same line as you a voc.\n"
"//\n"
"// The file is organised as:\n"
"// word or sentence in own language | word or sentence in foreign language\n"
"// another word in own language | with translation\n"
"//\n"
"// You can enter several translations of a single word in both lang1 and lang2\n"
"// Separate these with the '' sign.\n"
"// word 1  word 2 | trans 1  trans 2  trans 3\n"
"//\n"
"//\n"
"// You get the idea, right?\n"
"// As you can see I've decided to use the '|' and '' signs as delimeters.\n"
"// These are the default delimeters\n"
"//\n"
"// You can change all delimiters by adding special settings to your vocfiles\n"
"// or in a file called 'settings' in your .vocatest directory. You can also\n"
"// edit the path to the directory in which you wish to keep you vocfiles from\n"
"// within that file. The default vocfile directory is /vocfiles.\n"
"//\n"
"// Some words below. English | Swedish.\n"
"hello  hi | hej  hejsan \n"
"planet of the apes | apornas planet\n"
"pizza | pizza\n"
"chewing gum | tuggummi // this is a comment";

// Startup string
string startupString =
"Welcome to VocaTest 2.0 Special Windows Build - aka \"Andrea Edition\"\n"
"Author: Olof Holmgren - oloho217@student.liu.se\n"
"License: GPL";

// Declaring some functions
void programLoop();
void setGlobals(const string fileName);
void dispHelp();
void load(string file, bool append);
void listVocs();
void test(bool backw, bool strict);
void save(string &file);
void randomize();
void buildVoc(string &file);

bool isAnsInList(deque<string> &ansList, const string &ans, const bool &strict);
string testVoc(VocPair vp, const bool &backw, const unsigned int &pos);
bool checkAnswerAndDisplayText(deque<string> &ansList,
			       const string &answer, const bool &strict);
bool ifSettingApply(string &row);
bool wantsToOverwriteFile(const std::string &file);
void writeIfPossible(const string &file, const string &data);


// Declaring tha var's...
VocList vocList;
string baseDir = "";
string setting_langDelim = "|";
string setting_wordDelim = "";
string setting_vocDir;

// Da main thing
int main()
{
  // Unportable stuff
  setting_vocDir = "vocfiles/";
 
  if(!fileExists("settings.txt"))
    {
      cout << endl;
      cout << "=======================================================" << endl;
      cout << "Setting up VocaTest..." << endl;

      string settingString = "langDelim = \" | \"\n"
                             "wordDelim = \"  \"\n"
                             "vocDir = \"" + setting_vocDir + "\"";
      
      cout << "Creating file " << baseDir + "settings.txt";
      if(writeToFile(baseDir + "settings.txt", settingString))
	cout << "\t\t [DONE]" << endl;
      else
	cout << "\t\t [FAILED]" << endl;;
      
      cout << "Setup done" << endl;
      cout << "=======================================================" << endl;
    }
  // End of unportable stuff
  
  setGlobals(baseDir + "settings");
  programLoop();
  exit(0);
}

// Go through a given file for settings and apply these
void setGlobals(const string file)
{
  ifstream inFile(file.c_str());
  string row;
  while(getline(inFile, row))
    ifSettingApply(row);

  inFile.close();
  return;
}

// The main loop thing
void programLoop()
{
  cout << endl << startupString << endl;
  dispHelp();
  cout << "<VocaTest>";
  
  string input = "";
  while(getline(cin, input))
    {      
      string command = "";
      string arg = "";
      
      // Filter out the actual command and remove it from the input string
      rmLeadWs(input);
      rmTrailWs(input);
      if(input.find(' ') != string::npos)
	{
	  int spacePos = input.find(' ');
	  
	  command = input.substr(0, spacePos);
	  arg = input.substr(spacePos, input.length());
	}
      else // No space found ==> no args!
	{
	  command = input;
	  arg = "";
	}
      
      // Filter out the argument, if any
      rmLeadWs(arg);
      rmTrailWs(arg);
      
      // Execute the command!
      if(command == "load")
	load(arg, false);
      else if(command == "append")
	load(arg, true);
      else if(command == "list")
	listVocs();
      else if(command == "test")
	{
	  if(arg == "strict")
	    test(false, true);
	  else if(arg != "")
	    cout << "No such argument for that command. "
		 << "Type 'help' to see the available commands." << endl;
	  else
	    test(false, false);
	}
      else if(command == "testbackw")
	{
	   if(arg == "strict")
	     test(true, true);
	   else if(arg != "")
	     cout << "No such argument for that command. "
		  << "Type 'help' to see the available commands." << endl;
	   else
	     test(true, false);
	}
      else if(command == "randomize")
	randomize();
      else if(command == "buildvoc")
	buildVoc(arg);
      else if(command == "save")
	save(arg);
      else if(command == "help")
	dispHelp();
      else if(command == "$abort")
	cout << "That command can only be executed while running a test"
	     << endl;
      else if(command == "exit" || command == "quit")
	return;
      else
	cout << "No such command. Type 'help' to see the available commands"
	     << endl;
      
      cout << "<VocaTest>";
    }
}

// Print the help
void dispHelp()
{
  cout << helpString << endl << endl;;
}

// Randomize the list
void randomize()
{
  vocList.randomize();
  cout << "All tests will now be performed in random order" << endl;
}

// Build the sample vocfile
void buildVoc(string &file)
{
  if(file == "")
    cout << "And under what filename should I build this vocfile?" << endl;
  else
    {
      file = setting_vocDir + file;
      if(fileExists(file))
	{
	  if(wantsToOverwriteFile(file))
	    writeIfPossible(file, sampleFile);
	  else
	    cout << "File NOT saved" << endl;
	}
      else // There is no such file
	writeIfPossible(file, sampleFile);
    }
}

// Build the sample vocfile
void save(string &file)
{
  if(vocList.size() == 0)
    cout << "No vocs loaded. I don't save things that do not exist" << endl;
  else if(file == "")
    cout << "And under what filename should I save this file?" << endl;
  else
    {
      file = setting_vocDir + file;
      
      string data = "";
      for(unsigned int i = 0; i < vocList.size(); i++)
	{
	  data +=  stringDeque2String(vocList.getVoc(i).getQuestions(false),
					setting_wordDelim);
	  data += setting_langDelim;
	  data += stringDeque2String(vocList.getVoc(i).getAnswers(false),
				     setting_wordDelim);
	  data += "\n";
	}

      if(fileExists(file))
	{
	  if(wantsToOverwriteFile(file))
	    writeIfPossible(file, data);
	  else
	    cout << "File NOT saved" << endl;
	}
      else // file did not exist
	writeIfPossible(file, data);
    }
}

// Load a file
void load(string file, bool append)
{
  if(file == "")
    cout << "And what file am I supposed to read from?" << endl;
  else
    {
      if(!append)
	vocList.clear();
      
      file = setting_vocDir + file;
      
      if(fileExists(file))
      {
	ifstream inFile(file.c_str());

	string row;
	bool vocFound = false;
	while(getline(inFile, row))
	  {
	    if(!ifSettingApply(row))
	      {
		unsigned int pos = row.find(setting_langDelim);
		if((row.find(setting_langDelim) != string::npos) &&
		   (pos != row.length() - 1) && (pos != 0))
		  {
		    // Filter lang1 & lang2
		    string lang1W = row.substr(0, pos);
		    string lang2W = row.substr(pos + setting_langDelim.length(),
					       row.length());
		    
		    deque<string> lang1L = split(lang1W, setting_wordDelim);   
		    deque<string> lang2L = split(lang2W, setting_wordDelim);
		    
		    vocList.addVoc(VocPair(lang1L, lang2L));
		    vocFound = true;
		  }
	      }
	  }
	inFile.close();
	
	if(!vocFound)
	  cout << file << " contains no valid vocs" << endl;
	else
	  cout << "Loaded file: " << file << endl;
      }
      else
	cout << "\"" + file + "\"" << " can't be opened for reading" << endl;
    }
}

// Print the vocs
void listVocs()
{
  cout << endl;
  unsigned int i;
  for(i = 0; i < vocList.size(); i++)
    {
      string questions =
	stringDeque2String(vocList.getVoc(i).getQuestions(false), ", ");
      string answers =
	stringDeque2String(vocList.getVoc(i).getAnswers(false), ", ");
      
      cout << questions << " = " << answers << endl;
    }
  
  cout << endl << "Number of vocs: " << i << endl;
}

// Time to do a voctest...
void test(bool backw, bool strict)
{
  if(vocList.size() == 0)
    cout << "You have to load a vocfile first!" << endl;
  else
    {
      if(vocList.isRandom())
	vocList.randomize();
      
      bool abort = false;
      bool correct = false;
      unsigned int nrOfCorrect = 0;
      unsigned int i = 0;
      while(!abort && (i < vocList.size()))
	{
	  string answer = testVoc(vocList.getVoc(i), backw, i + 1);
	  
	  if(answer == "$abort") // Are we trying to abort?
	      abort = true;
	  else
	    {
	      deque<string> ansList = vocList.getVoc(i).getAnswers(backw);
	      
	      // checkAnswer... takes care of all the tasks associated with
	      // checking if one is correct. As a side effect, it also returns
	      // a bool (true if we were correct and false otherwise.
	      if(checkAnswerAndDisplayText(ansList, answer, strict))
		correct = true;

	      while(strict && (ansList.size() > 0) && !abort)
		{
		  answer = testVoc(vocList.getVoc(i), backw, i + 1);
		  if(answer == "$abort")
		    abort = true;
		  else
		    {
		      if(checkAnswerAndDisplayText(ansList, answer, strict))
			correct = true;
		    }
		}
	      if(correct)
		nrOfCorrect++;
	    }
	  correct = false;
	  i++;
	}
      
      if(abort)
	cout << "Test aborted" << endl;
      else
	{
	  cout << "Nr of correct: " << nrOfCorrect << endl;
	  cout << "Nr of errors: " << vocList.size() - nrOfCorrect << endl;
	  cout << "Total nr of vocs: " << vocList.size() << endl;
	}
    }
}

////////////////////////////////////////
// Helper functions below
////////////////////////////////////////

// Test a voc...
string testVoc(VocPair vp, const bool &backw, const unsigned int &pos)
{
  cout << vp.getQuestion(backw) << "\t\t (" << pos << "/"
       << vocList.size() << ")" << endl;

  string answer;
  getline(cin, answer);
  rmLeadWs(answer);
  rmTrailWs(answer);
  return answer;
}

// Checks a list for the answer. If we call it with strict = true it will
// delete the correct anser from the list. Otherwise, it will not.
// Returns a boolean representing whether the answer was found or not.
// Also displays all appropriate texts associated with these operations.
bool checkAnswerAndDisplayText(deque<string> &ansList,
			       const string &answer, const bool &strict)
{
  if(isAnsInList(ansList, answer, strict))
    {
      cout << "Correct!" << endl << endl;
      return true;
    }
  else
    {
      cout << "Wrong!" << endl
	   << "The correct answer was: "
	   << stringDeque2String(ansList, " or ")
	   << endl << endl;
      return false;
    }  
}

// Checks a list for the answer. If we call it with strict = true it will
// delete the correct anser from the list. Otherwise, it will not.
// Returns a boolean representing whether the answer was found or not.
bool isAnsInList(deque<string> &ansList, const string &ans, const bool &strict)
{
  unsigned int j = 0;
  while(j < ansList.size())
    {
      if(ans == ansList[j])
	{
	  if(strict)
	    {
	      deque<string> temp;
	      for(unsigned int i = 0; i < j; i++)
		temp.push_back(ansList[i]);
	      
	      for(unsigned int i = j; i < ansList.size() - 1; i++)
		temp.push_back(ansList[i + 1]);

	      ansList = temp;
	    }

	  return true;
	}
      else
	j++;
    }
  return false;
}

// If the string (row) is a setting, this will be applyed
// Returns a true if it was a setting and false if it was not
bool ifSettingApply(string &row)
{
  rmLeadWs(row);
  rmTrailWs(row);
  rmComments(row);
      
  unsigned int pos = row.find("=");
  if((row.find("=") != string::npos) && (pos != row.length() - 1) && (pos != 0))
    {
      // Filter lang1 & lang2
      string setting = row.substr(0, pos);
      string value = row.substr(pos + 1, row.length());
      
      rmTrailWs(setting);
      
      // Go from "arg" to just arg
      pos = value.find("\"");
      value = value.substr(pos + 1, value.length());
      pos = value.find("\"");
      value = value.substr(0, pos);
      
      if(setting == "langDelim")
	{
	  setting_langDelim = value;
	  return true;
	}
      else if(setting == "wordDelim")
	{
	  setting_wordDelim = value;
	  return true;
	}
      else if(setting == "vocDir")
	{
	  setting_vocDir = value;
	  return true;
	}
    }
  return false;
}

// Ask the user if he wants to overwrite a ceration file.
// Returns true for yes and false for no
bool wantsToOverwriteFile(const string &file)
{
  cout << "Overwrite file " << file << "? (y/n) ";
  
  bool retval = false;
  string input = "";
  while(!(input == "y" || input == "n"))
    {
      getline(cin, input);
      rmLeadWs(input);
      rmTrailWs(input);
      
      if(input == "y")
	retval = true;

      // else if == n do nothing
    }
  return retval;
}

// Write to the file if you can and display error messages otherwise
void writeIfPossible(const string &file, const string &data)
{
  if(writeToFile(file, sampleFile))
    cout << file << " saved" << endl;
  else
    cout << "Could not open file " << file << " for writing" << endl
	 << "File NOT saved" << endl;

  return;
}
